/**
 * WACKY LIB socket library
 *
 * Copyright (C) 2001-2002 Jeffrey Fulmer <jdfulmer@armstrong.com>
 * This file is part of libwacky
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifdef  HAVE_CONFIG_H
# include <config.h>
#endif/*HAVE_CONFIG_H*/

#include <sock.h>
#include <wacky.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#ifdef  HAVE_SYS_TIMES_H
# include <sys/times.h>
#endif/*HAVE_SYS_TIMES_H*/
 
#if  TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif/*TIME_WITH_SYS_TIME*/

/** 
 * local prototypes 
 */
int mknblock( int sock, int nonblock ); 
ssize_t socket_write( int sock, const void *vbuf, size_t len );  
ssize_t ssl_socket_write( CONN *C, const void *vbuf, size_t len );

/**
 * WLsocket
 * returns int, socket handle
 */
int
WLsocket( CONN *C, const char *hn, int port )
{
  int conn;
  int serr; 
  int herrno;
  struct linger      ling;
  struct hostent     *hp;
  struct hostent     hent;
#if defined(_AIX)
  char *aixbuf;
  int  rc;
#endif/*_AIX*/
  struct sockaddr_in cli; 
  char buf[1024]; 
  char hbf[8192];

  C->sock = -1;
#ifdef  HAVE_SSL
  C->ssl  = NULL;
  C->ctx  = NULL;
#endif/*HAVE_SSL*/
  
  if( C->prot == WACKYS ){
    #ifdef HAVE_SSL
      SSL_load_error_strings();
      SSLeay_add_ssl_algorithms();
      C->ctx = SSL_CTX_new( SSLv3_client_method());
      if( C->ctx == NULL ){ perror( "SSL: ctx is NULL" ); }
      /* http://www.openssl.org/support/faq.html#USER1
       * Perhaps someday I'll learn to read the FAQ
       * first and then code second, but probably not.
       * Not all OSes have /dev/urandom, we must seed
       * the PRNG
       */
      memset( buf, 0, sizeof( buf ));
      RAND_seed( buf, sizeof( buf ));
      C->ssl = SSL_new( C->ctx );
      SSL_set_connect_state( C->ssl );
    #else
      return -1;
    #endif /* HAVE_SSL */
  }

  /* create a socket, return -1 on failure */
  if(( C->sock = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ){
    perror( "WLsocket: failed to create socket" );
    return -1;
  }

#if defined(__GLIBC__)  
  /* for systems using GNU libc */
  if(( gethostbyname_r( hn, &hent, hbf, sizeof(hbf), &hp, &herrno ) < 0 )){
    hp = NULL;
  }
#elif defined(sun)
  /* Solaris 5++ */
  hp = gethostbyname_r( hn, &hent, hbf, sizeof(hbf), &herrno );
#elif defined(_AIX)
  aixbuf = (char*)xmalloc( 9000 );
  rc  = gethostbyname_r(hn, (struct hostent *)aixbuf,
                       (struct hostent_data *)(aixbuf + sizeof(struct hostent)));  
  hp = (struct hostent*)aixbuf; 
#elif ( defined(hpux) || defined(__osf__) )
  hp = gethostbyname( hn );
  herrno = h_errno;
#else 
  /* simply hoping that gethostbyname is thread-safe */
  hp = gethostbyname( hn );
  herrno = h_errno; 
#endif/*OS SPECIFICS*/

  /**
   * gethostbyname_r error handling 
   * and assignment
   */
  if( !hp ){
    perror( "WLsocket: hostname failure" );
    return -1;
  }
  memset((void*) &cli, 0, sizeof( cli ));
  memcpy( &cli.sin_addr, hp->h_addr, hp->h_length );
  memcpy( &cli.sin_addr, hp->h_addr, hp->h_length );
  cli.sin_family = AF_INET;
  cli.sin_port = htons( port );

  ling.l_onoff = 1;
  ling.l_linger = 0;
  if( setsockopt( C->sock, SOL_SOCKET, SO_LINGER, &ling, sizeof( ling )) < 0 ){ 
    perror( "WLsocket: socket option error" );
    return -1;    
  } 

  /* connect to the host */ 
  if((conn = connect(C->sock, (struct sockaddr *)&cli, sizeof(struct sockaddr_in)))<0){
    switch( errno ){
    case EACCES:
      perror( "EACCES");
      WLclose( C );
      return( -1 );
    case EADDRNOTAVAIL:
      perror( "EADDRNOTAVAIL");
      WLclose( C );
      return( -1 );
    case ETIMEDOUT:
      perror( "ETIMEDOUT");
      WLclose( C );
      return( -1 );
    case ECONNREFUSED:
      perror( "ECONNREFUSED");
      WLclose( C );
      return( -1 );
    case ENETUNREACH:
      perror( "ENETUNREACH");
      WLclose( C );
      return( -1 );
    default:
      perror("DEFAULT: connection failed");
      WLclose( C );
      return -1;
    } 
    perror( "connection failed" );
    WLclose( C );
    return -1;
  }
#if 0
  else if( errno == EINPROGRESS ){
    /** 
     * FIX ME: once we switch to non-blocking i-o,
     * we will have to wait until the socket is 
     * writeable to continue. we can ignore this now... 
     */
    perror( "WLsocket: connecting..." );
    return -1;
  }
#endif 

  /* if requested, encrypt the transaction  */
  if( C->prot == WACKYS ){
    #ifdef HAVE_SSL
      /* currently a fatal error, should it be? */
      if(( SSL_set_fd( C->ssl, C->sock )) < 0 ){
        perror( "unable to create secure socket!" );
        exit( 1 );
      }
      if(( serr = SSL_connect( C->ssl )) < 0 ){
        int problem = SSL_get_error( C->ssl, serr );
        fprintf( stderr, "SSL_connect: want to %s more...\n",
               ( problem == SSL_ERROR_WANT_READ) ? "read" : "write");
        return -1;
      }
      SSL_get_cipher( C->ssl );
    #else
      return -1;
    #endif /* HAVE_SSL */
  }
   
  return( C->sock );
}

/**
 * local function
 * set socket to non-blocking
 */
int
mknblock( int sock, int nonblock )
{
#if HAVE_FCNTL_H
  int flags;
  int retval;
 
  flags = fcntl( sock, F_GETFL, 0 );
  if( flags < 0 ){
    perror("fcntl");
    return -1;
  }
  if( nonblock ){
    flags |=  O_NDELAY;
  }
  else{
    flags &= ~O_NDELAY;
  }
  retval = fcntl( sock, F_SETFL, flags);
  if( retval < 0 ){
    perror("fcntl");
    return -1;
  }
#elif defined( FIONBIO )
  ioctl_t status;
 
  status = nb ? 1 : 0;
  return ioctl( sd, FIONBIO, &status );
#endif
} 

int
WLsocket_check( CONN *C, SDSET test )
{
  int res;
  struct timeval timeout;
 
  FD_ZERO(&C->rset);
  FD_ZERO(&C->wset);
  FD_SET( C->sock, &C->rset );
  FD_SET( C->sock, &C->wset );
  memset((void *)&timeout, '\0', sizeof( struct timeval ));
 
  if(( mknblock( C->sock, TRUE )) < 0 ){
    perror( "SIMBOTsocket: unable to set socket to non-blocking." );
    return -1;
  }
 
  timeout.tv_sec  = 30;
  timeout.tv_usec = 0;
 
  switch( test ){
  case READ:
    if(( res = select( C->sock+1, &C->rset, NULL, NULL, &timeout )) < 1 ){
      close( C->sock );
      return -1;
    }
    break;
  case WRITE:
    if(( res = select( C->sock+1, NULL, &C->wset, NULL, &timeout )) < 1 ){
      close( C->sock );
      return -1;
    }
    break;
  case RDWR:
    if(( res = select( C->sock+1, &C->rset, &C->wset, NULL, &timeout )) < 1 ){
      close( C->sock );
      return -1;
    }
    break;
  }
 
  if(( mknblock( C->sock, FALSE )) < 0 ){
    perror( "WLsocket: unable to set socket to non-blocking." );
    return -1;
  }
  FD_CLR( C->sock, &C->rset );
  
  return 0;
}

/**
 * local function
 * returns ssize_t
 * writes vbuf to sock
 */
ssize_t
socket_write( int sock, const void *vbuf, size_t len )
{
  size_t      n;
  ssize_t     w;
  const char *buf;
 
  buf = vbuf;
  n   = len;
  while( n > 0 ){
    if(( w = write( sock, buf, n )) <= 0 ){
      if( errno == EINTR )
        w = 0;
      else
        return( -1 );
    }
    n    -= w;
    buf += n;
  }
  return( len );
}

/**
 * local function
 * returns ssize_t
 * writes vbuf to sock
 */
ssize_t
ssl_socket_write( CONN *C, const void *vbuf, size_t len )
{
  size_t      n;
  ssize_t     w;
  const char *buf;

#ifdef HAVE_SSL
  buf = vbuf;
  n   = len;
  while( n > 0 ){
    if(( w = SSL_write( C->ssl, buf, n )) <= 0 ){
      if( errno == EINTR )
        w = 0;
      else
        return( -1 );
    }
    n    -= w;
    buf += n;
  }
  return( len );
#else
  perror( "protocol not supported" );
  return -1;
#endif/*HAVE_SSL*/
}


/**
 * returns int
 * takes sock as a parameter to localize the
 * handler for thread safety.
 */
ssize_t
WLread_byte( CONN *C, void *buf, size_t len )
{
  ssize_t n;
  if( C->prot == WACKYS ){
  #ifdef  HAVE_SSL
    if(( n = SSL_read( C->ssl, buf, len )) < 0 ){
      perror( "WLread_byte: read error %d", C->sock );
    } 
  #endif/*HAVE_SSL*/ 
  }
  else{
    if(( n = read( C->sock, buf, len )) < 0 ){
      perror( "WLread_byte: read error" ); 
    }
  }
  return( n ); 
}

ssize_t
WLsocket_read( CONN *C, void *vbuf, size_t len )
{
  size_t      n;
  ssize_t     r;
  char *buf;
 
  buf = vbuf;
  n   = len;
  
  if( C->prot == WACKYS ){
  #ifdef HAVE_SSL
    while( n > 0 ){ 
      if(( r = SSL_read( C->ssl, buf, n )) < 0 ){
        if( errno == EINTR )
          r = 0;
        else
          return( -1 );
      }
      else if( r == 0 ) break; 
      n   -= r;
      buf += r;
    }   /* end of while    */
  #endif/*HAVE_SSL*/
  }  
  else{
    while( n > 0 ){
      if(( r = read( C->sock, buf, n )) < 0 ){
        if( errno == EINTR )
          r = 0;
        else
          return( -1 );
      }
      else if( r == 0 ) break;
      n   -= r;
      buf += r;
    } /* end of while */ 
  }   /* end of else  */
  return( len - n );  
}  

/**
 * returns void
 * socket_write wrapper function.
 */
void
WLsocket_write( CONN *C, const void *buf, size_t len )
{
  int bytes;

  if( C->prot == WACKYS ){
    /* handle WACKYS protocol */
    #ifdef HAVE_SSL
    if(( bytes = ssl_socket_write( C, buf, len )) != len ){
      perror( "WLssl_socket_write: ERROR" );
    }
    #else
      perror( "WLssl_socket_write: protocol NOT supported" );
    #endif/*HAVE_SSL*/
  }
  else{
    /* assume WACKY */
    if(( bytes = socket_write( C->sock, buf, len )) != len ){
      perror( "WLsocket_write: ERROR" );
    }
  }
  return;
} 

/**
 * returns void
 * frees ssl resources if using ssl and
 * closes the connection and the socket.
 */
void
WLclose( CONN *C )
{
  #ifdef  HAVE_SSL
    if( C->prot == WACKYS ){
      SSL_shutdown( C->ssl );
    }
    SSL_free( C->ssl );
    SSL_CTX_free( C->ctx );
  #endif/*HAVE_SSL*/
  close( C->sock ); 
  return;
} 


